@@ -65,6 +65,7 @@ uhm_server_error_quark (void)
 }
 
 static void uhm_server_dispose (GObject *object);
+static void uhm_server_finalize (GObject *object);
 static void uhm_server_get_property (GObject *object, guint property_id, GValue *value, GParamSpec *pspec);
 static void uhm_server_set_property (GObject *object, guint property_id, const GValue *value, GParamSpec *pspec);
 
@@ -78,14 +79,20 @@ static void load_file_iteration_thread_cb (GTask *task, gpointer source_object,
 static GFileInputStream *load_file_stream (GFile *trace_file, GCancellable *cancellable, GError **error);
 static SoupMessage *load_file_iteration (GFileInputStream *input_stream, SoupURI *base_uri, GCancellable *cancellable, GError **error);
 
+static void apply_expected_domain_names (UhmServer *self);
+
 struct _UhmServerPrivate {
 	/* UhmServer is based around HTTP/HTTPS, and cannot be extended to support other application-layer protocols.
 	 * If libuhttpmock is extended to support other protocols (e.g. IMAP) in future, a new UhmImapServer should be
 	 * created. It may be possible to share code with UhmServer, but the APIs for the two would be sufficiently
-	 * different to not be mergeable. */
+	 * different to not be mergeable.
+	 *
+	 * The SoupServer is *not* thread safe, so all calls to it must be marshalled through the server_context, which
+	 * it runs in. */
 	SoupServer *server;
 	UhmResolver *resolver;
 	GThread *server_thread;
+	GMainContext *server_context;
 
 	/* TLS certificate. */
 	GTlsCertificate *tls_certificate;
@@ -94,6 +101,9 @@ struct _UhmServerPrivate {
 	SoupAddress *address; /* unowned */
 	guint port;
 
+	/* Expected resolver domain names. */
+	gchar **expected_domain_names;
+
 	GFile *trace_file;
 	GFileInputStream *input_stream;
 	GFileOutputStream *output_stream;
@@ -144,6 +154,7 @@ uhm_server_class_init (UhmServerClass *klass)
 	gobject_class->get_property = uhm_server_get_property;
 	gobject_class->set_property = uhm_server_set_property;
 	gobject_class->dispose = uhm_server_dispose;
+	gobject_class->finalize = uhm_server_finalize;
 
 	klass->handle_message = real_handle_message;
 	klass->compare_messages = real_compare_messages;
@@ -324,6 +335,7 @@ uhm_server_dispose (GObject *object)
 
 	g_clear_object (&priv->resolver);
 	g_clear_object (&priv->server);
+	g_clear_pointer (&priv->server_context, g_main_context_unref);
 	g_clear_object (&priv->input_stream);
 	g_clear_object (&priv->trace_file);
 	g_clear_object (&priv->input_stream);
@@ -338,6 +350,17 @@ uhm_server_dispose (GObject *object)
 	G_OBJECT_CLASS (uhm_server_parent_class)->dispose (object);
 }
 
+static void
+uhm_server_finalize (GObject *object)
+{
+	UhmServerPrivate *priv = UHM_SERVER (object)->priv;
+
+	g_strfreev (priv->expected_domain_names);
+
+	/* Chain up to the parent class */
+	G_OBJECT_CLASS (uhm_server_parent_class)->finalize (object);
+}
+
 static void
 uhm_server_get_property (GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
 {
@@ -752,7 +775,7 @@ error:
 static SoupMessage *
 trace_to_soup_message (const gchar *trace, SoupURI *base_uri)
 {
-	SoupMessage *message;
+	SoupMessage *message = NULL;
 	const gchar *i, *j, *method;
 	gchar *uri_string, *response_message;
 	SoupHTTPVersion http_version;
@@ -828,6 +851,7 @@ trace_to_soup_message (const gchar *trace, SoupURI *base_uri)
 		trace += strlen ("HTTP/1.0");
 	} else {
 		g_warning ("Unrecognised HTTP version ‘%s’.", trace);
+		http_version = SOUP_HTTP_1_1;
 	}
 
 	if (*trace != '\n') {
@@ -909,7 +933,7 @@ trace_to_soup_message (const gchar *trace, SoupURI *base_uri)
 	return message;
 
 error:
-	g_object_unref (message);
+	g_clear_object (&message);
 
 	return NULL;
 }
@@ -1099,7 +1123,8 @@ uhm_server_unload_trace (UhmServer *self)
  *
  * Loading the trace file may be cancelled from another thread using @cancellable.
  *
- * On error, @error will be set and the state of the #UhmServer will not change.
+ * On error, @error will be set and the state of the #UhmServer will not change. A #GIOError will be set if there is
+ * a problem reading the trace file.
  *
  * Since: 0.1.0
  */
@@ -1225,6 +1250,7 @@ uhm_server_load_trace_async (UhmServer *self, GFile *trace_file, GCancellable *c
  * Finishes an asynchronous operation started by uhm_server_load_trace_async().
  *
  * On error, @error will be set and the state of the #UhmServer will not change.
+ * See uhm_server_load_trace() for details on the error domains used.
  *
  * Since: 0.1.0
  */
@@ -1242,15 +1268,31 @@ uhm_server_load_trace_finish (UhmServer *self, GAsyncResult *result, GError **er
 	self->priv->received_message_state = UNKNOWN;
 }
 
+/* Must only be called in the server thread. */
+static gboolean
+server_thread_quit_cb (gpointer user_data)
+{
+	UhmServer *self = user_data;
+	UhmServerPrivate *priv = self->priv;
+
+	soup_server_quit (priv->server);
+
+	return G_SOURCE_REMOVE;
+}
+
 static gpointer
 server_thread_cb (gpointer user_data)
 {
 	UhmServer *self = user_data;
 	UhmServerPrivate *priv = self->priv;
 
-	/* Run the server. */
+	g_main_context_push_thread_default (priv->server_context);
+
+	/* Run the server. This will create a main loop and iterate the server_context until soup_server_quit() is called. */
 	soup_server_run (priv->server);
 
+	g_main_context_pop_thread_default (priv->server_context);
+
 	return NULL;
 }
 
@@ -1273,9 +1315,11 @@ void
 uhm_server_run (UhmServer *self)
 {
 	UhmServerPrivate *priv = self->priv;
-	struct sockaddr_in sock;
+	union {
+		struct sockaddr sock;
+		struct sockaddr_in in;
+	} sock;
 	SoupAddress *addr;
-	GMainContext *thread_context;
 
 	g_return_if_fail (UHM_IS_SERVER (self));
 	g_return_if_fail (priv->resolver == NULL);
@@ -1283,24 +1327,23 @@ uhm_server_run (UhmServer *self)
 
 	/* Grab a loopback IP to use. */
 	memset (&sock, 0, sizeof (sock));
-	sock.sin_family = AF_INET;
-	sock.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
-	sock.sin_port = htons (0); /* random port */
+	sock.in.sin_family = AF_INET;
+	sock.in.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
+	sock.in.sin_port = htons (0); /* random port */
 
-	addr = soup_address_new_from_sockaddr ((struct sockaddr *) &sock, sizeof (sock));
+	addr = soup_address_new_from_sockaddr (&sock.sock, sizeof (sock));
 	g_assert (addr != NULL);
 
 	/* Set up the server. If (priv->tls_certificate != NULL) it will be a HTTPS server;
 	 * otherwise it will be a HTTP server. */
-	thread_context = g_main_context_new ();
+	priv->server_context = g_main_context_new ();
 	priv->server = soup_server_new ("interface", addr,
 	                                "tls-certificate", priv->tls_certificate,
-	                                "async-context", thread_context,
+	                                "async-context", priv->server_context,
 	                                "raw-paths", TRUE,
 	                                NULL);
 	soup_server_add_handler (priv->server, "/", server_handler_cb, self, NULL);
 
-	g_main_context_unref (thread_context);
 	g_object_unref (addr);
 
 	/* Grab the randomly selected address and port. */
@@ -1309,10 +1352,13 @@ uhm_server_run (UhmServer *self)
 
 	/* Set up the resolver. It is expected that callers will grab the resolver (by calling uhm_server_get_resolver())
 	 * immediately after this function returns, and add some expected hostnames by calling uhm_resolver_add_A() one or
-	 * more times, before starting the next test. */
+	 * more times, before starting the next test.Or they could call uhm_server_set_expected_domain_names() any time. */
 	priv->resolver = uhm_resolver_new ();
 	g_resolver_set_default (G_RESOLVER (priv->resolver));
 
+	/* Note: This must be called before notify::resolver, so the user can add extra domain names in that callback if desired. */
+	apply_expected_domain_names (self);
+
 	g_object_freeze_notify (G_OBJECT (self));
 	g_object_notify (G_OBJECT (self), "address");
 	g_object_notify (G_OBJECT (self), "port");
@@ -1340,17 +1386,22 @@ void
 uhm_server_stop (UhmServer *self)
 {
 	UhmServerPrivate *priv = self->priv;
+	GSource *idle;
 
 	g_return_if_fail (UHM_IS_SERVER (self));
 	g_return_if_fail (priv->server != NULL);
 	g_return_if_fail (priv->resolver != NULL);
 
 	/* Stop the server. */
-	soup_server_disconnect (priv->server);
+	idle = g_idle_source_new ();
+	g_source_set_callback (idle, server_thread_quit_cb, self, NULL);
+	g_source_attach (idle, priv->server_context);
+
 	g_thread_join (priv->server_thread);
 	priv->server_thread = NULL;
 	uhm_resolver_reset (priv->resolver);
 
+	g_clear_pointer (&priv->server_context, g_main_context_unref);
 	g_clear_object (&priv->server);
 	g_clear_object (&priv->resolver);
 
@@ -1420,7 +1471,8 @@ uhm_server_set_trace_directory (UhmServer *self, GFile *trace_directory)
  *
  * This function has undefined behaviour if #UhmServer:trace-directory is %NULL.
  *
- * On failure, @error will be set and the #UhmServer state will remain unchanged.
+ * On failure, @error will be set and the #UhmServer state will remain unchanged. See uhm_server_start_trace_full() for
+ * details of the error domains used.
  *
  * Since: 0.1.0
  */
@@ -1455,7 +1507,9 @@ uhm_server_start_trace (UhmServer *self, const gchar *trace_name, GError **error
  * If #UhmServer:enable-online is %FALSE, the given @trace_file is loaded using uhm_server_load_trace() and then a mock server is
  * started using uhm_server_run().
  *
- * On failure, @error will be set and the #UhmServer state will remain unchanged.
+ * On failure, @error will be set and the #UhmServer state will remain unchanged. A #GIOError will be set if logging is enabled
+ * (#UhmServer:enable-logging) and there is a problem writing to the trace file; or if a trace needs to be loaded and there is a problem
+ * reading from the trace file.
  *
  * Since: 0.1.0
  */
@@ -1648,6 +1702,10 @@ uhm_server_set_enable_logging (UhmServer *self, gboolean enable_logging)
  * comparison mode and the received message chunk corresponds to an unexpected message in the trace file, a %UHM_SERVER_ERROR will
  * be returned.
  *
+ * <note><para>In common cases where message log data only needs to be passed to a #UhmServer and not (for example) logged to an
+ * application-specific file or the command line as  well, it is simpler to use uhm_server_received_message_chunk_from_soup(), passing
+ * it directly to soup_logger_set_printer(). See the documentation for uhm_server_received_message_chunk_from_soup() for details.</para></note>
+ *
  * Since: 0.1.0
  */
 void
@@ -1772,6 +1830,99 @@ uhm_server_received_message_chunk (UhmServer *self, const gchar *message_chunk,
 	}
 }
 
+/**
+ * uhm_server_received_message_chunk_with_direction:
+ * @self: a #UhmServer
+ * @direction: single character indicating the direction of message transmission
+ * @data: single line of a message which was received
+ * @data_length: length of @data in bytes
+ * @error: (allow-none): return location for a #GError, or %NULL
+ *
+ * Convenience version of uhm_server_received_message_chunk() which takes the
+ * message @direction and @data separately, as provided by libsoup in a
+ * #SoupLoggerPrinter callback.
+ *
+ * <informalexample><programlisting>
+ * UhmServer *mock_server;
+ * SoupSession *session;
+ * SoupLogger *logger;
+ *
+ * static void
+ * soup_log_printer (SoupLogger *logger, SoupLoggerLogLevel level, char direction, const char *data, gpointer user_data)
+ * {
+ * 	/<!-- -->* Pass the data to libuhttpmock. *<!-- -->/
+ *	UhmServer *mock_server = UHM_SERVER (user_data);
+ * 	uhm_server_received_message_chunk_with_direction (mock_server, direction, data, strlen (data), NULL);
+ * }
+ *
+ * mock_server = uhm_server_new ();
+ * session = soup_session_new ();
+ *
+ * logger = soup_logger_new (SOUP_LOGGER_LOG_BODY, -1);
+ * soup_logger_set_printer (logger, (SoupLoggerPrinter) soup_log_printer, g_object_ref (mock_server), g_object_unref);
+ * soup_session_add_feature (session, SOUP_SESSION_FEATURE (logger));
+ * g_object_unref (logger);
+ *
+ * /<!-- -->* Do something with mock_server here. *<!-- -->/
+ * </programlisting></informalexample>
+ *
+ * Since: 0.3.0
+ */
+void
+uhm_server_received_message_chunk_with_direction (UhmServer *self, char direction, const gchar *data, goffset data_length, GError **error)
+{
+	gchar *message_chunk;
+
+	g_return_if_fail (UHM_IS_SERVER (self));
+	g_return_if_fail (direction == '<' || direction == '>' || direction == ' ');
+	g_return_if_fail (data != NULL);
+	g_return_if_fail (data_length >= -1);
+	g_return_if_fail (error == NULL || *error == NULL);
+
+	/* This is inefficient and not nul-safe, but it’ll do for now. */
+	message_chunk = g_strdup_printf ("%c %s", direction, data);
+	uhm_server_received_message_chunk (self, message_chunk, (data_length > -1) ? data_length + 2 : -1, error);
+	g_free (message_chunk);
+}
+
+/**
+ * uhm_server_received_message_chunk_from_soup:
+ * @logger: a #SoupLogger
+ * @level: the detail level of this log message
+ * @direction: the transmission direction of the message
+ * @data: message data
+ * @user_data: (allow-none): user data passed to the #SoupLogger, or %NULL
+ *
+ * Convenience version of uhm_server_received_message_chunk() which can be passed directly to soup_logger_set_printer()
+ * to forward all libsoup traffic logging to a #UhmServer. The #UhmServer must be passed to soup_logger_set_printer() as
+ * its user data.
+ *
+ * <informalexample><programlisting>
+ * UhmServer *mock_server;
+ * SoupSession *session;
+ * SoupLogger *logger;
+ *
+ * mock_server = uhm_server_new ();
+ * session = soup_session_new ();
+ *
+ * logger = soup_logger_new (SOUP_LOGGER_LOG_BODY, -1);
+ * soup_logger_set_printer (logger, uhm_server_received_message_chunk_from_soup, g_object_ref (mock_server), g_object_unref);
+ * soup_session_add_feature (session, SOUP_SESSION_FEATURE (logger));
+ * g_object_unref (logger);
+ *
+ * /<!-- -->* Do something with mock_server here. *<!-- -->/
+ * </programlisting></informalexample>
+ *
+ * Since: 0.3.0
+ */
+void
+uhm_server_received_message_chunk_from_soup (SoupLogger *logger, SoupLoggerLogLevel level, char direction, const char *data, gpointer user_data)
+{
+	/* Deliberately don’t do strict validation of parameters here, since we can’t be entirely sure what libsoup throws our way. */
+	UhmServer *mock_server = UHM_SERVER (user_data);
+	uhm_server_received_message_chunk_with_direction (mock_server, direction, data, strlen (data), NULL);
+}
+
 /**
  * uhm_server_get_address:
  * @self: a #UhmServer
@@ -1907,3 +2058,59 @@ uhm_server_set_default_tls_certificate (UhmServer *self)
 
 	return cert;
 }
+
+static void
+apply_expected_domain_names (UhmServer *self)
+{
+	UhmServerPrivate *priv = self->priv;
+	const gchar *ip_address;
+	guint i;
+
+	if (priv->resolver == NULL) {
+		return;
+	}
+
+	uhm_resolver_reset (priv->resolver);
+
+	if (priv->expected_domain_names == NULL) {
+		return;
+	}
+
+	ip_address = uhm_server_get_address (self);
+	g_assert (ip_address != NULL);
+
+	for (i = 0; priv->expected_domain_names[i] != NULL; i++) {
+		uhm_resolver_add_A (priv->resolver, priv->expected_domain_names[i], ip_address);
+	}
+}
+
+/**
+ * uhm_server_set_expected_domain_names:
+ * @self: a #UhmServer
+ * @domain_names: (array zero-terminated=1) (allow-none) (element-type utf8): %NULL-terminated array of domain names to expect, or %NULL to not expect any
+ *
+ * Set the domain names which are expected to have requests made of them by the client code interacting with this #UhmServer.
+ * This is a convenience method which calls uhm_resolver_add_A() on the server’s #UhmResolver for each of the domain names
+ * listed in @domain_names. It associates them with the server’s current IP address, and automatically updates the mappings
+ * if the IP address or resolver change.
+ *
+ * Note that this will reset all records on the server’s #UhmResolver, replacing all of them with the provided @domain_names.
+ *
+ * It is safe to add further domain names to the #UhmResolver in a callback for the #GObject::notify signal for #UhmServer:resolver;
+ * that signal is emitted after the resolver is cleared and these @domain_names are added.
+ *
+ * Since: 0.3.0
+ */
+void
+uhm_server_set_expected_domain_names (UhmServer *self, const gchar * const *domain_names)
+{
+	gchar **new_domain_names;
+
+	g_return_if_fail (UHM_IS_SERVER (self));
+
+	new_domain_names = g_strdupv ((gchar **) domain_names);  /* may be NULL */
+	g_strfreev (self->priv->expected_domain_names);
+	self->priv->expected_domain_names = new_domain_names;
+
+	apply_expected_domain_names (self);
+}
